"""
Gathers GPU utilization status via 'nvgpu' library from multiple hosts
and publishes in via a web app.

This web app can act either as an agent (just publishing status of GPUs on one
machine in JSON via REST API) or master (in addition aggregating statuses from
all machines in a cluster and showing them on a HTML page). Master instance can
also act as an agent.

Master:
NVGPU_CLUSTER_CFG=$(pwd)/nvgpu_master.cfg FLASK_APP=nvgpu.webapp flask run --host 0.0.0.0 --port 1080

Agent:
FLASK_APP=nvgpu.webapp flask run --host 0.0.0.0 --port 1080
"""

from ansi2html import Ansi2HTMLConverter
from flask import Flask
from flask import Response
from flask_restful import Resource, Api

from nvgpu.agent import machine_status
from nvgpu.master import gather_reports, format_reports_to_ansi

app = Flask(__name__)
api = Api(app)

app.config.from_envvar('NVGPU_CLUSTER_CFG', silent=True)

# agent is either a URL to another machine or 'self' for direct access
# if no agents specified it will only report its own status
agents = app.config.get('AGENTS', ['self'])


class GpuStatus(Resource):
    def get(self):
        return machine_status()


class GpuClusterStatus(Resource):
    def get(self):
        return gather_reports(agents)


class GpuClusterStatusHTML(Resource):
    def get(self):
        reports = gather_reports(agents)
        ansi_text = format_reports_to_ansi(agents, reports)
        html = ansi_text_to_html(ansi_text)
        return Response(html, mimetype='text/html')


def ansi_text_to_html(text):
    conv = Ansi2HTMLConverter(scheme='osx', dark_bg=False)
    html = conv.convert(text)
    html = html.replace(
        '.body_background { background-color: #AAAAAA; }',
        '.body_background { background-color: #FFFFFF; }')
    return html


# Possibly publish this only if NVML is available on this host.
api.add_resource(GpuStatus, '/gpu_status')
api.add_resource(GpuClusterStatusHTML, '/')
api.add_resource(GpuClusterStatus, '/cluster_status')

if __name__ == '__main__':
    app.run(debug=True, port=1080)



import requests
from requests import ReadTimeout, ConnectionError

from nvgpu.agent import machine_status
from nvgpu.list_gpus import format_table, device_table


def format_reports_to_ansi(hosts, reports):
    def format_report(host, report):
        if 'error' not in report:
            return '\n'.join(['%s (%s)' % (host, report['hostname']), format_table(device_table(report['gpus']))])
        else:
            return '\n'.join([host, report['error']])

    return '\n\n'.join([
        format_report(host, report)
        for host, report in zip(hosts, reports)])


def gather_reports(hosts):
    return [download_report(host) for host in hosts]


def download_report(host):
    if host == 'self':
        # Avoid making a nested HTTP request to itself (may cause a deadlock
        # or timeout if the web app runs in a single thread).
        return machine_status()

    try:
        resp = requests.get(host + '/gpu_status', timeout=5)
        if resp.status_code != 200:
            return {'error': 'Bad response (%s: %s)' % (resp.status_code, resp.reason)}
    except ConnectionError:
        return {'error': 'Host not available'}
    except ReadTimeout:
        return {'error': 'Connection timeout'}

    return resp.json()



import re
import six
import subprocess


def gpu_info():
    gpus = [line for line in _run_cmd(['nvidia-smi', '-L']) if line]
    gpu_infos = [re.match('GPU ([0-9]+): (.+?) \(UUID: ([^)]+)\)', gpu) for gpu in gpus]
    gpu_infos = [info.groups() for info in gpu_infos if info is not None]
    gpu_infos = [dict(zip(['index', 'type', 'uuid'], info)) for info in gpu_infos]
    gpu_count = len(gpus)

    lines = _run_cmd(['nvidia-smi'])
    cuda_version = float(lines[2].split('CUDA Version: ')[1].split(' ')[0])
    if cuda_version < 11:
        line_distance = 3
        selected_lines = lines[7:7 + line_distance * gpu_count]
    else:
        line_distance = 4
        selected_lines = lines[8:8 + line_distance * gpu_count]
    for i in range(gpu_count):
        mem_used, mem_total = [int(m.strip().replace('MiB', '')) for m in
                               selected_lines[line_distance * i + 1].split('|')[2].strip().split('/')]
        gpu_infos[i]['mem_used'] = mem_used
        gpu_infos[i]['mem_total'] = mem_total
        gpu_infos[i]['mem_used_percent'] = 100. * mem_used / mem_total

    return gpu_infos


def _run_cmd(cmd):
    output = subprocess.check_output(cmd)
    if isinstance(output, bytes):
        output = output.decode('UTF-8')
    return output.split('\n')

def available_gpus(max_used_percent=20.):
    return [gpu['index'] for gpu in gpu_info() if gpu['mem_used_percent'] <= max_used_percent]



from __future__ import print_function

import arrow
import pandas as pd
import psutil
import pynvml as nv
import six
from tabulate import tabulate
from termcolor import colored

from nvgpu.nvml import nvml_context


def device_status(device_index):
    handle = nv.nvmlDeviceGetHandleByIndex(device_index)
    device_name = nv.nvmlDeviceGetName(handle)
    if isinstance(device_name, bytes):
        device_name = device_name.decode('UTF-8')
    nv_procs = nv.nvmlDeviceGetComputeRunningProcesses(handle)
    try:
        utilization = nv.nvmlDeviceGetUtilizationRates(handle).gpu
    except nv.NVMLError:
        utilization = None

    if six.PY3:
        clock_mhz = nv.nvmlDeviceGetClockInfo(handle, nv.NVML_CLOCK_SM)
    else:
        # old API in nvidia-ml-py
        clock_mhz = nv.nvmlDeviceGetClock(handle, nv.NVML_CLOCK_SM, 0)
    temperature = nv.nvmlDeviceGetTemperature(handle, nv.NVML_TEMPERATURE_GPU)
    memory = nv.nvmlDeviceGetMemoryInfo(handle)
    pids = []
    users = []
    dates = []
    cmd = None
    for nv_proc in nv_procs:
        pid = nv_proc.pid
        pids.append(pid)
        try:
            proc = psutil.Process(pid)
            users.append(proc.username())
            dates.append(proc.create_time())
            if cmd is None:
                cmd = parse_cmd_roughly(proc.cmdline())
        except psutil.NoSuchProcess:
            users.append('?')
    return {
        'type': device_name,
        'is_available': len(pids) == 0,
        'pids': ','.join([str(pid) for pid in pids]),
        'users': ','.join(users),
        'running_since': arrow.get(min(dates)).humanize() if len(dates) > 0 else None,
        'utilization': utilization,
        'memory': memory,
        'clock_mhz': clock_mhz,
        'temperature': temperature,
        'cmd': cmd,
    }


def parse_cmd_roughly(args):
    cmdline = ' '.join(args)
    if 'python -m ipykernel_launcher' in cmdline:
        return 'jupyter'
    python_script = [arg for arg in args if arg.endswith('.py')]
    if len(python_script) > 0:
        return python_script[0]
    else:
        return cmdline if len(cmdline) <= 25 else cmdline[:25] + '...'


def device_statuses():
    with nvml_context():
        device_count = nv.nvmlDeviceGetCount()
        return [device_status(device_index) for device_index in range(device_count)]


def device_table(rows):
    df = pd.DataFrame(rows, columns=[
        'is_available', 'type', 'utilization', 'memory', 'clock_mhz', 'temperature', 'users', 'running_since', 'pids', 'cmd'
    ])
    return df


def format_table(df):
    def make_status(row):
        if row['users'] == '?':
            return '[!]'
        elif row['is_available']:
            return '[ ]'
        else:
            return '[~]'

    def color_by_status(status):
        if status == '[ ]':
            return 'green'
        elif status == '[~]':
            return 'blue'
        elif status == '[!]':
            return 'red'
        else:
            raise ValueError(status)
    df['status'] = df.apply(make_status, axis=1)
    df['color'] = df['status'].apply(color_by_status)
    df['util.'] = df['utilization'].apply(lambda u: '%03s %%' % u if u is not None else 'N/A')
    df['mem. used'] = df['memory'].apply(lambda m: str(round(m.used / 1000000000., 2)) + "GB")
    df['MHz'] = df['clock_mhz']
    df['temp.'] = df['temperature']
    df['since'] = df['running_since']
    for col in ['since', 'cmd']:
        df[col] = df[col].apply(lambda v: v if v is not None else '')
    cols = ['status', 'type', 'util.', 'mem. used', 'temp.', 'MHz', 'users', 'since', 'pids', 'cmd']
    for col in cols:
        df[col] = [colored(row[col], row['color']) for i, row in df.iterrows()]
    df = df[[col for col in cols if col not in ['color']]]
    return tabulate(df, headers='keys')


def pretty_list_gpus():
    rows = device_statuses()
    df = device_table(rows)
    print(format_table(df))


if __name__ == '__main__':
    pretty_list_gpus()



from setuptools import setup


def read_file(path):
    with open(path, 'r') as f:
        return f.read()


setup(name='nvgpu',
    version='0.10.0',
    description='NVIDIA GPU tools',
    url='https://github.com/rossumai/nvgpu',
    author='Bohumir Zamecnik, Rossum',
    author_email='bohumir.zamecnik@rossum.ai',
    license='MIT',
    packages=['nvgpu'],
    zip_safe=False,
    install_requires=[
        'ansi2html',
        'arrow',
        'flask',
        'flask_restful',
        'nvidia-ml-py; python_version < "3"',
        'nvidia-ml-py3; python_version >= "3" and python_version < "3.6"',
        'pynvml; python_version >= "3.6"',
        'pandas',
        'psutil',
        'requests',
        'six',
        'termcolor',
        'tabulate',
    ],
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',

        'Operating System :: POSIX :: Linux',
    ],
    entry_points={
        'console_scripts': [
            'nvgpu = nvgpu.__main__:main',
            'nvl = nvgpu.list_gpus:pretty_list_gpus'
        ]
    }, )
